@@ -1,6 +1,6 @@
 /* DAT.C */
 
-/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2001      */
+/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2003      */
 
 #include "hercules.h"
 
@@ -14,12 +14,16 @@
 
 #if !defined(_GEN_ARCH)
 
-#define  _GEN_ARCH 390
-#include "dat.c"
+#if defined(_ARCHMODE2)
+ #define  _GEN_ARCH _ARCHMODE2
+ #include "dat.c"
+#endif
 
-#undef   _GEN_ARCH
-#define  _GEN_ARCH 370
-#include "dat.c"
+#if defined(_ARCHMODE3)
+ #undef   _GEN_ARCH
+ #define  _GEN_ARCH _ARCHMODE3
+ #include "dat.c"
+#endif
 
 #endif /*!defined(_GEN_ARCH)*/
 
