@@ -1,8 +1,9 @@
-/* FEAT390.H	(c) Copyright Jan Jaeger, 2000-2001		     */
-/*		ESA/390 feature definitions			     */
+/* FEAT390.H    (c) Copyright Jan Jaeger, 2000-2003          */
+/*      ESA/390 feature definitions              */
 
 /* This file MUST NOT contain #undef statements */
-#define FEATURE_ESAME_INSTALLED
+#if defined(OPTION_390_MODE)
+#define _ARCH_390_NAME "ESA/390"
 
 #define FEATURE_4K_STORAGE_KEYS
 #define FEATURE_ACCESS_REGISTERS
@@ -22,9 +23,12 @@
 #define FEATURE_DUAL_ADDRESS_SPACE
 #define FEATURE_EMULATE_VM
 #define FEATURE_EXPANDED_STORAGE
+#define FEATURE_EXPEDITED_SIE_SUBSET  
 #define FEATURE_EXTENDED_STORAGE_KEYS
 #define FEATURE_EXTENDED_TOD_CLOCK
 #define FEATURE_EXTENDED_TRANSLATION
+#define FEATURE_EXTENDED_TRANSLATION_FACILITY_2
+#define FEATURE_EXTERNAL_INTERRUPT_ASSIST
 #define FEATURE_FETCH_PROTECTION_OVERRIDE
 // #define FEATURE_FPS_EXTENSIONS
 #define FEATURE_HERCULES_DIAGCALLS
@@ -33,6 +37,7 @@
 #define FEATURE_HYPERVISOR
 #define FEATURE_IMMEDIATE_AND_RELATIVE
 #define FEATURE_INCORRECT_LENGTH_INDICATION_SUPPRESSION
+#define FEATURE_IO_ASSIST
 #define FEATURE_INTERPRETIVE_EXECUTION
 #define FEATURE_LOCK_PAGE
 #define FEATURE_LINKAGE_STACK
@@ -42,8 +47,11 @@
 #define FEATURE_MVS_ASSIST
 #define FEATURE_PAGE_PROTECTION
 #define FEATURE_PERFORM_LOCKED_OPERATION
+#define FEATURE_PER
 #define FEATURE_PER2
 #define FEATURE_PRIVATE_SPACE
+#define FEATURE_PROTECTION_INTERCEPTION_CONTROL
+#define FEATURE_REGION_RELOCATE
 #define FEATURE_RESUME_PROGRAM
 #define FEATURE_S390_DAT
 #define FEATURE_SERVICE_PROCESSOR
@@ -55,7 +63,11 @@
 #define FEATURE_SUBSPACE_GROUP
 #define FEATURE_SUPPRESSION_ON_PROTECTION
 #define FEATURE_SYSTEM_CONSOLE
+#define FEATURE_TEST_BLOCK
 #define FEATURE_TRACING
+#define FEATURE_WAITSTATE_ASSIST
+
 // #define FEATURE_VECTOR_FACILITY
 
+#endif /*defined(OPTION_390_MODE)*/
 /* end of FEAT390.H */
