@@ -1,5 +1,7 @@
-/* FEAT900.H	(c) Copyright Jan Jaeger, 2000-2001		     */
-/*		ESAME feature definitions			     */
+/* FEAT900.H    (c) Copyright Jan Jaeger, 2000-2003          */
+/*      ESAME feature definitions                */
+#if defined(OPTION_900_MODE)
+#define _ARCH_900_NAME "ESAME"
 
 /* This file MUST NOT contain #undef statements */
 #define FEATURE_4K_STORAGE_KEYS
@@ -21,10 +23,12 @@
 #define FEATURE_EMULATE_VM
 #define FEATURE_ESAME
 #define FEATURE_EXPANDED_STORAGE
+#define FEATURE_EXPEDITED_SIE_SUBSET  
 #define FEATURE_EXTENDED_STORAGE_KEYS
 #define FEATURE_EXTENDED_TOD_CLOCK
 #define FEATURE_EXTENDED_TRANSLATION
 #define FEATURE_EXTENDED_TRANSLATION_FACILITY_2
+#define FEATURE_EXTERNAL_INTERRUPT_ASSIST
 #define FEATURE_FETCH_PROTECTION_OVERRIDE
 // #define FEATURE_FPS_EXTENSIONS
 #define FEATURE_HERCULES_DIAGCALLS
@@ -34,6 +38,7 @@
 #define FEATURE_IMMEDIATE_AND_RELATIVE
 #define FEATURE_INCORRECT_LENGTH_INDICATION_SUPPRESSION
 #define FEATURE_INTERPRETIVE_EXECUTION
+#define FEATURE_IO_ASSIST
 #define FEATURE_LOAD_REVERSED
 #define FEATURE_LOCK_PAGE
 #define FEATURE_LINKAGE_STACK
@@ -41,9 +46,12 @@
 #define FEATURE_MVS_ASSIST
 #define FEATURE_PAGE_PROTECTION
 #define FEATURE_PERFORM_LOCKED_OPERATION
+#define FEATURE_PER
 #define FEATURE_PER2
 #define FEATURE_PRIVATE_SPACE
+#define FEATURE_PROTECTION_INTERCEPTION_CONTROL
 #define FEATURE_RESUME_PROGRAM
+#define FEATURE_REGION_RELOCATE
 #define FEATURE_SERVICE_PROCESSOR
 #define FEATURE_CHSC
 #define FEATURE_SQUARE_ROOT
@@ -53,6 +61,9 @@
 #define FEATURE_SUBSPACE_GROUP
 #define FEATURE_SUPPRESSION_ON_PROTECTION
 #define FEATURE_SYSTEM_CONSOLE
+#define FEATURE_TEST_BLOCK
 #define FEATURE_TRACING
+#define FEATURE_WAITSTATE_ASSIST
 
+#endif /*defined(OPTION_900_MODE)*/
 /* end of FEAT900.H */
