@@ -1,4 +1,4 @@
-/* FEATALL.H    (c) Copyright Jan Jaeger, 2000-2001                  */
+/* FEATALL.H    (c) Copyright Jan Jaeger, 2000-2003                  */
 /*              Architecture-dependent macro definitions             */
 
 /*-------------------------------------------------------------------*/
@@ -7,6 +7,10 @@
 /*-------------------------------------------------------------------*/
 #define MAX_CPU_ENGINES               2 /* Maximum number of engines */
 #undef  SMP_SERIALIZATION               /* Serialize storage for SMP */
+#define OPTION_370_MODE                 /* Generate S/370 support    */
+#define OPTION_390_MODE                 /* Generate ESA/390 support  */
+#define OPTION_900_MODE                 /* Generate ESAME support    */
+#define OPTION_LPP_RESTRICT             /* Disable Licensed Software */
 #define VECTOR_SECTION_SIZE         128 /* Vector section size       */
 #define VECTOR_PARTIAL_SUM_NUMBER     1 /* Vector partial sum number */
 #define CKD_MAXFILES                  4 /* Max files per CKD volume  */
@@ -29,28 +33,35 @@
 #define OPTION_FAST_LOGICAL             /* Performance option        */
 #define OPTION_FAST_PREFIX              /* Performance option        */
 #define OPTION_REDUCED_INVAL            /* Performance option        */
-#define OPTION_FETCHIBYTE               /* Performance option        */
-#undef  OPTION_CS_USLEEP                /* Sleep if CS collision     */
 #define OPTION_FAST_INTCOND             /* Ea CPU has intcond COND   */
-#define OPTION_IOINTQ                   /* I/O interrupt queue       */
 #define OPTION_SYNCIO                   /* Synchronous I/O option    */
-#undef  OPTION_IODELAY         /*  1000 ** IO delay in uSec          */
+#define OPTION_IODELAY_KLUDGE           /* IODELAY kludge for linux  */
+#define OPTION_IODELAY_LINUX_DEFAULT 800/* Default if OSTAILOR LINUX */
 #undef  OPTION_FOOTPRINT_BUFFER /* 2048 ** Size must be a power of 2 */
 #undef  OPTION_INSTRUCTION_COUNTING     /* First use trace and count */
 #define OPTION_CKD_KEY_TRACING          /* Trace CKD search keys     */
 #undef  OPTION_CMPSC_DEBUGLVL      /* 3 ** 1=Exp 2=Comp 3=Both debug */
-#undef  MODEL_DEPENDENT_CS              /* CS, CDS, CSP always store */
+#ifdef  WIN32
+#define OPTION_SELECT_KLUDGE         10 /* fd's to reserve for select*/
+#endif
 #undef  MODEL_DEPENDENT_STCM            /* STCM, STCMH always store  */
+#define OPTION_NOP_MODEL158_DIAGNOSE    /* NOP mod 158 specific diags*/
+#define FEATURE_ALD_FORMAT            0
+#define FEATURE_SIE_MAXZONES          8
 
-/* This is set as appropriate by configure and makefile.w32; don't
-   define or undefine it here. -mdz */
-/* #undef NO_SIGABEND_HANDLER */
+#define OPTION_HTTP_SERVER              /* HTTP server support       */
 
 
-/* The following option is set as appropriate by configure and
-   makefile.w32; don't define or undefine it here. */
-/* #define OPTION_FTHREADS */           /* Fish pthreads replacement */
-
+/* Allow for compiler command line overrides */
+#if defined(OPTION_370_MODE) && defined(NO_370_MODE)
+ #undef OPTION_370_MODE
+#endif
+#if defined(OPTION_390_MODE) && defined(NO_390_MODE)
+ #undef OPTION_390_MODE
+#endif
+#if defined(OPTION_900_MODE) && defined(NO_900_MODE)
+ #undef OPTION_900_MODE
+#endif
 
 /* OPTION_FISHIO only possible with OPTION_FTHREADS */
 #if defined(OPTION_FTHREADS)
@@ -60,9 +71,14 @@
 #endif
 
 
-#define FEATURE_ALD_FORMAT            0
+/* CTCI-W32 only valid for Win32 */
+#if defined(WIN32)
+#define OPTION_W32_CTCI
+#endif // defined(WIN32)
+
 
 #undef FEATURE_4K_STORAGE_KEYS
+#undef FEATURE_2K_STORAGE_KEYS
 #undef FEATURE_ACCESS_REGISTERS
 #undef FEATURE_ADDRESS_LIMIT_CHECKING
 #undef FEATURE_BASIC_FP_EXTENSIONS
@@ -75,6 +91,7 @@
 #undef FEATURE_CALLED_SPACE_IDENTIFICATION
 #undef FEATURE_CANCEL_IO_FACILITY
 #undef FEATURE_CHANNEL_SUBSYSTEM
+#undef FEATURE_CHANNEL_SWITCHING
 #undef FEATURE_CHECKSUM_INSTRUCTION
 #undef FEATURE_COMPARE_AND_MOVE_EXTENDED
 #undef FEATURE_COMPRESSION
@@ -82,13 +99,14 @@
 #undef FEATURE_DUAL_ADDRESS_SPACE
 #undef FEATURE_EMULATE_VM
 #undef FEATURE_ESAME
-#undef FEATURE_ESAME_INSTALLED
 #undef FEATURE_ESAME_N3_ESA390
 #undef FEATURE_EXPANDED_STORAGE
+#undef FEATURE_EXPEDITED_SIE_SUBSET  
 #undef FEATURE_EXTENDED_STORAGE_KEYS
 #undef FEATURE_EXTENDED_TOD_CLOCK
 #undef FEATURE_EXTENDED_TRANSLATION
 #undef FEATURE_EXTENDED_TRANSLATION_FACILITY_2
+#undef FEATURE_EXTERNAL_INTERRUPT_ASSIST
 #undef FEATURE_FETCH_PROTECTION_OVERRIDE
 #undef FEATURE_FPS_EXTENSIONS
 #undef FEATURE_HERCULES_DIAGCALLS
@@ -99,6 +117,7 @@
 #undef FEATURE_INCORRECT_LENGTH_INDICATION_SUPPRESSION
 #undef FEATURE_INTERPRETIVE_EXECUTION
 #undef FEATURE_INTERVAL_TIMER
+#undef FEATURE_IO_ASSIST
 #undef FEATURE_LINKAGE_STACK
 #undef FEATURE_LOAD_REVERSED
 #undef FEATURE_LOCK_PAGE
@@ -108,13 +127,18 @@
 #undef FEATURE_MVS_ASSIST
 #undef FEATURE_PAGE_PROTECTION
 #undef FEATURE_PERFORM_LOCKED_OPERATION
+#undef FEATURE_PER
 #undef FEATURE_PER2
 #undef FEATURE_PRIVATE_SPACE
+#undef FEATURE_PROTECTION_INTERCEPTION_CONTROL
+#undef FEATURE_REGION_RELOCATE
 #undef FEATURE_RESUME_PROGRAM
 #undef FEATURE_S370_CHANNEL
 #undef FEATURE_S390_DAT
+#undef FEATURE_S370E_EXTENDED_ADDRESSING
 #undef FEATURE_SERVICE_PROCESSOR
 #undef FEATURE_SEGMENT_PROTECTION
+#undef FEATURE_CHANNEL_SWITCHING
 #undef FEATURE_CHSC
 #undef FEATURE_SQUARE_ROOT
 #undef FEATURE_STORAGE_KEY_ASSIST
@@ -123,7 +147,9 @@
 #undef FEATURE_SUBSPACE_GROUP
 #undef FEATURE_SUPPRESSION_ON_PROTECTION
 #undef FEATURE_SYSTEM_CONSOLE
+#undef FEATURE_TEST_BLOCK
 #undef FEATURE_TRACING
 #undef FEATURE_VECTOR_FACILITY
+#undef FEATURE_WAITSTATE_ASSIST
 
 /* end of FEATALL.H */
