@@ -1,11 +1,11 @@
-/* FEATCHK.H	(c) Copyright Jan Jaeger, 2000-2001		     */
-/*		Feature definition consistency checks		     */
+/* FEATCHK.H    (c) Copyright Jan Jaeger, 2000-2003          */
+/*      Feature definition consistency checks            */
 
 /*-------------------------------------------------------------------*/
-/* Perform various checks on feature combinations, and set	     */
-/* additional flags to percolate certain features such as	     */
-/* SIE down to lower architecture levels such that these	     */
-/* can include emulation support				     */
+/* Perform various checks on feature combinations, and set       */
+/* additional flags to percolate certain features such as        */
+/* SIE down to lower architecture levels such that these         */
+/* can include emulation support                     */
 /*-------------------------------------------------------------------*/
 
 #if defined(FEATCHK_CHECK_ALL)
@@ -20,6 +20,9 @@
  #if defined(FEATURE_ESAME)
   #define _FEATURE_ZSIE
  #endif
+ #if defined(FEATURE_PROTECTION_INTERCEPTION_CONTROL)
+  #define _FEATURE_PROTECTION_INTERCEPTION_CONTROL
+ #endif
 #endif
 
 /* _FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE is used for host 
@@ -30,8 +33,7 @@
  #define _FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE
 #endif
 
-#if defined(FEATURE_2K_STORAGE_KEYS) \
- && defined(FEATURE_4K_STORAGE_KEYS)
+#if defined(FEATURE_2K_STORAGE_KEYS)
  #define _FEATURE_2K_STORAGE_KEYS
 #endif
 
@@ -47,6 +49,10 @@
  #define _FEATURE_CHANNEL_SUBSYSTEM
 #endif
 
+#if defined(FEATURE_SYSTEM_CONSOLE)
+ #define _FEATURE_SYSTEM_CONSOLE
+#endif
+
 #if defined(FEATURE_EXPANDED_STORAGE)
  #define _FEATURE_EXPANDED_STORAGE
 #endif
@@ -59,6 +65,34 @@
  #define _FEATURE_CPU_RECONFIG
 #endif
 
+#if defined(FEATURE_PER)
+ #define _FEATURE_PER
+#endif
+
+#if defined(FEATURE_PER2)
+ #define _FEATURE_PER2
+#endif
+
+#if defined(FEATURE_EXPEDITED_SIE_SUBSET)
+#define _FEATURE_EXPEDITED_SIE_SUBSET
+#endif
+
+#if defined(FEATURE_REGION_RELOCATE)
+ #define _FEATURE_REGION_RELOCATE
+#endif
+
+#if defined(FEATURE_IO_ASSIST)
+ #define _FEATURE_IO_ASSIST
+#endif
+
+#if defined(FEATURE_WAITSTATE_ASSIST)
+ #define _FEATURE_WAITSTATE_ASSIST
+#endif
+
+#if defined(FEATURE_EXTERNAL_INTERRUPT_ASSIST)
+ #define _FEATURE_EXTERNAL_INTERRUPT_ASSIST
+#endif
+
 #undef _VSTORE_C_STATIC
 #if !defined(OPTION_NO_INLINE_VSTORE)
  #define _VSTORE_C_STATIC static inline
@@ -87,27 +121,87 @@
  #define _LOGICAL_C_STATIC
 #endif
 
+#if !defined(OPTION_370_MODE) \
+  && !defined(OPTION_390_MODE) \
+  && !defined(OPTION_900_MODE)
+ #error No Architecture mode
+#endif
+#if defined(OPTION_370_MODE)
+ #define _370
+ #define _ARCHMODE1 370
+ #define ARCH_370 0
+#endif
+
+#if defined(OPTION_390_MODE)
+ #define _390
+ #if !defined(_ARCHMODE1)
+  #define _ARCHMODE1 390
+  #define ARCH_390 0
+ #else
+  #define _ARCHMODE2 390
+  #define ARCH_390 1
+ #endif
+#endif
+
+#if defined(OPTION_900_MODE)
+ #define _900
+ #if !defined(_ARCHMODE2)
+  #define _ARCHMODE2 900
+  #define ARCH_900 1
+ #else
+  #define _ARCHMODE3 900
+  #define ARCH_900 2
+ #endif
+#endif
+
+#if !defined(ARCH_370)
+ #define ARCH_370 -1
+#endif
+#if !defined(ARCH_390)
+ #define ARCH_390 -1
+#endif
+#if !defined(ARCH_900)
+ #define ARCH_900 -1
+#endif
+
+#if defined(_ARCHMODE3)
+ #define GEN_MAXARCH    3+2
+#elif defined(_ARCHMODE2)
+ #define GEN_MAXARCH    2+2
+#else 
+ #define GEN_MAXARCH    1+2
+#endif
+
+#if defined(_900) && !defined(_390)
+ #error OPTION_390_MODE must be enabled for OPTION_900_MODE
+#endif
+
 #else /*!defined(FEATCHK_CHECK_ALL)*/
 
 /* When ESAME is installed then all instructions
    marked N3 in the reference are also available
    in ESA/390 mode */
-#if defined(FEATURE_ESAME_INSTALLED)
+#if defined(_900)
  #define FEATURE_ESAME_N3_ESA390
-#endif /*defined(FEATURE_ESAME_INSTALLED)*/
+#endif /*defined(_900)*/
 
 #if !defined(FEATURE_2K_STORAGE_KEYS) \
  && !defined(FEATURE_4K_STORAGE_KEYS)
  #error Storage keys must be 2K or 4K
 #endif
 
+#if defined(FEATURE_EXTENDED_STORAGE_KEYS)
+ #if !defined(FEATURE_S370E_EXTENDED_ADDRESSING)
+  #define FEATURE_S370E_EXTENDED_ADDRESSING
+ #endif
+#endif
+
 #if defined(FEATURE_EXPANDED_STORAGE) \
  && !defined(FEATURE_4K_STORAGE_KEYS)
  #error Expanded storage cannot be defined with 2K storage keys
 #endif
 
-#if ( defined(FEATURE_ESAME_INSTALLED) || defined(FEATURE_ESAME) ) \
- && defined(FEATURE_VECTOR_FACILITY)
+#if defined(_900) && defined(FEATURE_VECTOR_FACILITY)
  #error Vector Facility not supported on ESAME capable processors
 #endif
  
@@ -116,11 +210,22 @@
  #error Move page facility cannot be defined with 2K storage keys
 #endif
 
+#if defined(FEATURE_ESAME) \
+  && defined(FEATURE_INTERPRETIVE_EXECUTION) \
+  && !defined(_FEATURE_SIE)
+ #error ESA/390 SIE must be defined when defining ESAME SIE
+#endif
+
 #if defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE) \
  && !defined(FEATURE_INTERPRETIVE_EXECUTION)
  #error MCDS only supported with SIE
 #endif
 
+#if defined(FEATURE_PROTECTION_INTERCEPTION_CONTROL) \
+ && !defined(FEATURE_INTERPRETIVE_EXECUTION)
+ #error Protection Interception Control only supported with SIE
+#endif
+
 #if defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE) \
  && !defined(FEATURE_STORAGE_KEY_ASSIST)
  #error MCDS requires storage key assist
@@ -136,6 +241,26 @@
  #error Storage Key assist only supported with SIE
 #endif
 
+#if defined(FEATURE_REGION_RELOCATE) \
+ && !defined(FEATURE_INTERPRETIVE_EXECUTION)
+ #error Region Relocate Facility only supported with SIE
+#endif
+
+#if defined(FEATURE_IO_ASSIST) \
+ && !defined(_FEATURE_SIE)
+ #error I/O Assist Feature only supported with SIE
+#endif
+
+#if defined(FEATURE_EXTERNAL_INTERRUPT_ASSIST) \
+ && !defined(_FEATURE_SIE)
+ #error External Interruption assist only supported with SIE
+#endif
+
+#if defined(FEATURE_EXPEDITED_SIE_SUBSET) \
+ && !defined(_FEATURE_SIE)
+ #error Expedited SIE Subset only supported with SIE
+#endif
+
 #if defined(FEATURE_ESAME) \
  && defined(FEATURE_VECTOR_FACILITY)
  #error Vector Facility not supported in ESAME mode
@@ -159,6 +284,11 @@
  #endif
 #endif
 
+#if defined(FEATURE_PER2) && !defined(FEATURE_PER)
+ #error FEATURE_PER must be defined when using FEATURE_PER2
+#endif
+
+
 #endif /*!defined(FEATALL_CHECKALL)*/
 
 /* end of FEATCHK.H */
