@@ -1,11 +1,11 @@
-/* GENERAL2.C   (c) Copyright Roger Bowler, 1994-2001                */
+/* GENERAL2.C   (c) Copyright Roger Bowler, 1994-2003                */
 /*              ESA/390 CPU Emulator                                 */
 /*              Instructions N-Z                                     */
 
 /*              (c) Copyright Peter Kuschnerus, 1999 (UPT & CFC)     */
 
-/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2001      */
-/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2001      */
+/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2003      */
+/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2003      */
 
 /*-------------------------------------------------------------------*/
 /* This module implements all general instructions of the            */
@@ -136,16 +136,16 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
         npa2 = LOGICAL_TO_ABS (npv2, b2, regs, ACCTYPE_READ, akey);
 
     /* all operands and page crossers valid, now alter ref & chg bits */
-    STORAGE_KEY(abs1) |= (STORKEY_REF | STORKEY_CHANGE);
+    STORAGE_KEY(abs1, regs) |= (STORKEY_REF | STORKEY_CHANGE);
     if (npa1)
-        STORAGE_KEY(npa1) |= (STORKEY_REF | STORKEY_CHANGE);
+        STORAGE_KEY(npa1, regs) |= (STORKEY_REF | STORKEY_CHANGE);
 
     /* Process operands from left to right */
     for ( i = 0; i <= l; i++ )
     {
         /* Fetch a byte from each operand */
-        byte1 = sysblk.mainstor[abs1];
-        byte2 = sysblk.mainstor[abs2];
+        byte1 = regs->mainstor[abs1];
+        byte2 = regs->mainstor[abs2];
 
         /* OR operand 1 with operand 2 */
         byte1 |= byte2;
@@ -154,7 +154,7 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
         if (byte1 != 0) cc = 1;
 
         /* Store the result in the destination operand */
-        sysblk.mainstor[abs1] = byte1;
+        regs->mainstor[abs1] = byte1;
 
         /* Increment first operand address */
         effective_addr1++;
@@ -423,16 +423,8 @@ VADR    effective_addr2,
         /* Release main-storage access lock */
         RELEASE_MAINLOCK(regs);
 
-#if MAX_CPU_ENGINES > 1 && defined(OPTION_CS_USLEEP)
-        /* It this is a failed locked operation
-           and there is more then 1 CPU in the configuration
-           and there is no broadcast synchronization in progress
-           then call the hypervisor to end this timeslice,
-           this to prevent this virtual CPU monopolizing
-           the physical CPU on a spinlock */
         if(regs->psw.cc && sysblk.numcpu > 1)
-            usleep(1L);
-#endif /* MAX_CPU_ENGINES > 1 && defined(OPTION_CS_USLEEP) */
+            sched_yield();
 
     }
 }
@@ -549,7 +541,7 @@ int     b2;                             /* effective address base    */
 VADR    effective_addr2;                /* effective address         */
 U32     n;                              /* 32-bit operand values     */
 U64     dreg;                           /* Double register work area */
-int     h, i, j, m;                     /* Integer work areas        */
+U32     h, i, j, m;                     /* Integer work areas        */
 
     RS(inst, execflag, regs, r1, r3, b2, effective_addr2);
 
@@ -631,7 +623,7 @@ int     r1, r3;                         /* Register numbers          */
 int     b2;                             /* effective address base    */
 VADR    effective_addr2;                /* effective address         */
 U32     n, n1, n2;                      /* 32-bit operand values     */
-int     i, j;                           /* Integer work areas        */
+U32     i, j;                           /* Integer work areas        */
 
     RS(inst, execflag, regs, r1, r3, b2, effective_addr2);
 
@@ -1194,6 +1186,7 @@ BYTE    i;                              /* Instruction byte 1        */
 PSA    *psa;                            /* -> prefixed storage area  */
 RADR    px;                             /* prefix                    */
 int     rc;                             /* Return code               */
+int     new_ilc;                        /* New ilc to set            */
 
     RR_SVC(inst, execflag, regs, i);
 
@@ -1213,7 +1206,7 @@ int     rc;                             /* Return code               */
     SIE_TRANSLATE(&px, ACCTYPE_WRITE, regs);
 
     /* Set the main storage reference and change bits */
-    STORAGE_KEY(px) |= (STORKEY_REF | STORKEY_CHANGE);
+    STORAGE_KEY(px, regs) |= (STORKEY_REF | STORKEY_CHANGE);
 
     /* Use the I-byte to set the SVC interruption code */
     regs->psw.intcode = i;
@@ -1227,7 +1220,7 @@ int     rc;                             /* Return code               */
 #endif
 
     /* Point to PSA in main storage */
-    psa = (void*)(sysblk.mainstor + px);
+    psa = (void*)(regs->mainstor + px);
 
 #if defined(FEATURE_BCMODE)
     /* For ECMODE, store SVC interrupt code at PSA+X'88' */
@@ -1240,14 +1233,33 @@ int     rc;                             /* Return code               */
         psa->svcint[3] = i;
     }
 
+    /* If there is a pending PER event and the PER addr is not
+       pointing to this SVC, then it must be EX to SVC case and
+       the new ILC needs to be set to 4 */
+
+#if defined(FEATURE_PER)
+    if (IS_IC_PER_IF(regs) && (regs->peradr != (regs->psw.IA - 2)))
+         new_ilc = 4;
+    else
+#endif /*defined(FEATURE_PER)*/
+         new_ilc = 2;
+
     /* Store current PSW at PSA+X'20' */
     ARCH_DEP(store_psw) ( regs, psa->svcold );
 
     /* Load new PSW from PSA+X'60' */
+    obtain_lock(&sysblk.intlock);
     rc = ARCH_DEP(load_psw) ( regs, psa->svcnew );
+    release_lock(&sysblk.intlock);
     if ( rc )
         ARCH_DEP(program_interrupt) (regs, rc);
 
+    /* load_psw() has set the ILC to zero.  This needs to
+       be reset to 2 for an eventual PER event
+       OR set to 4 if this is EX to SVC case */
+
+    regs->psw.ilc = new_ilc;
+
     /* Perform serialization and checkpoint synchronization */
     PERFORM_SERIALIZATION (regs);
     PERFORM_CHKPT_SYNC (regs);
@@ -1264,52 +1276,49 @@ DEF_INST(test_and_set)
 {
 int     b2;                             /* Base of effective addr    */
 VADR    effective_addr2;                /* Effective address         */
-RADR    abs;                            /* Absolute address          */
-BYTE    obyte;                          /* Operand byte              */
+RADR    abs2;                           /* Real address              */
+BYTE    old;                            /* Old value                 */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
     /* Perform serialization before starting operation */
     PERFORM_SERIALIZATION (regs);
 
-    abs = LOGICAL_TO_ABS (effective_addr2, b2, regs,
-					ACCTYPE_WRITE, regs->psw.pkey);
+    /* Get operand absolute address */
+    abs2 = LOGICAL_TO_ABS (effective_addr2, b2, regs,
+                           ACCTYPE_WRITE, regs->psw.pkey);
 
     /* Obtain main-storage access lock */
     OBTAIN_MAINLOCK(regs);
 
-    /* Fetch byte from operand address */
-    obyte = sysblk.mainstor[abs];
+    /* Get old value */
+    old = regs->mainstor[abs2];
 
-    /* Set all bits of operand to ones */
-    if(obyte != 0xFF)
-        sysblk.mainstor[abs] = 0xFF;
+    /* Attempt to exchange the values */
+    while (cmpxchg1(&old, 255, regs->mainstor + abs2));
+    regs->psw.cc = old >> 7;
 
     /* Release main-storage access lock */
     RELEASE_MAINLOCK(regs);
 
-    /* Set condition code from leftmost bit of operand byte */
-    regs->psw.cc = obyte >> 7;
-
     /* Perform serialization after completing operation */
     PERFORM_SERIALIZATION (regs);
 
+    if (regs->psw.cc == 1)
+    {
 #if defined(_FEATURE_SIE)
-    if(regs->sie_state && (regs->siebk->ic[0] & SIE_IC0_TS1))
-        longjmp(regs->progjmp, SIE_INTERCEPT_INST);
+        if((regs->sie_state && (regs->siebk->ic[0] & SIE_IC0_TS1)))
+        {
+            if( !OPEN_IC_PERINT(regs) )
+                longjmp(regs->progjmp, SIE_INTERCEPT_INST);
+            else
+                longjmp(regs->progjmp, SIE_INTERCEPT_INSTCOMP);
+        }
+        else
 #endif /*defined(_FEATURE_SIE)*/
-
-#if MAX_CPU_ENGINES > 1 && defined(OPTION_CS_USLEEP)
-        /* It this is a failed locked operation
-           and there is more then 1 CPU in the configuration
-           and there is no broadcast synchronization in progress
-           then call the hypervisor to end this timeslice,
-           this to prevent this virtual CPU monopolizing
-           the physical CPU on a spinlock */
-        if(regs->psw.cc && sysblk.numcpu > 1)
-            usleep(1L);
-#endif /* MAX_CPU_ENGINES > 1 && defined(OPTION_CS_USLEEP) */
-
+            if (sysblk.numcpu > 1)
+                sched_yield();
+    }
 }
 
 
@@ -1672,7 +1681,7 @@ BYTE    lbyte;                          /* Left result byte of pair  */
 
 /*-------------------------------------------------------------------*/
 /* 0102 UPT   - Update Tree                                      [E] */
-/*              (c) Copyright Peter Kuschnerus, 1999-2001            */
+/*              (c) Copyright Peter Kuschnerus, 1999-2003            */
 /* 64BIT INCOMPLETE                                                  */
 /*-------------------------------------------------------------------*/
 DEF_INST(update_tree)
@@ -1686,6 +1695,8 @@ int     ar1 = 4;                        /* Access register number    */
 
     E(inst, execflag, regs);
 
+    UNREFERENCED(inst);
+
     /* Check GR4, GR5 doubleword alligned */
     if ( regs->GR_L(4) & 0x00000007 || regs->GR_L(5) & 0x00000007 )
         ARCH_DEP(program_interrupt) (regs, PGM_SPECIFICATION_EXCEPTION);
@@ -1723,8 +1734,7 @@ int     ar1 = 4;                        /* Access register number    */
             /* Load GR2 and GR3 from tempword2 and tempword3 */
             regs->GR_L(2) = tempword2;
             regs->GR_L(3) = tempword3;
-
-           regs->GR_L(5) = tempword1;
+            regs->GR_L(5) = tempword1;
             cc = 0;
             break;
         }
@@ -1748,11 +1758,15 @@ int     ar1 = 4;                        /* Access register number    */
 
 #if !defined(_GEN_ARCH)
 
-#define  _GEN_ARCH 390
-#include "general2.c"
+#if defined(_ARCHMODE2)
+ #define  _GEN_ARCH _ARCHMODE2
+ #include "general2.c"
+#endif
 
-#undef   _GEN_ARCH
-#define  _GEN_ARCH 370
-#include "general2.c"
+#if defined(_ARCHMODE3)
+ #undef   _GEN_ARCH
+ #define  _GEN_ARCH _ARCHMODE3
+ #include "general2.c"
+#endif
 
 #endif /*!defined(_GEN_ARCH)*/
