@@ -1,7 +1,7 @@
 /*
 || ----------------------------------------------------------------------------
 ||
-|| HETGET.C     (c) Copyright Leland Lucius, 2000-2001
+|| HETGET.C     (c) Copyright Leland Lucius, 2000-2003
 ||              Released under terms of the Q Public License.
 ||
 || Extract files from an HET file
@@ -40,6 +40,7 @@ struct
 }
 opts = 
 {
+    NULL,
     NULL,
     0,
     0,
@@ -122,7 +123,7 @@ int reclen = 0;
 || Special flag to indicate whether or not we're being
 || run under the control of the external GUI facility.
 */
-static int extgui = 0;
+int extgui = 0;
 /*
 || Previously reported file position
 */
@@ -153,7 +154,7 @@ merge( SLLABEL *lab )
     if( opts.recfm == 0 )
     {
         opts.recfm = O_UNDEFINED;
-        for( i = 0 ; i < VALFMCNT ; i++ )
+        for( i = 0 ; i < (int)VALFMCNT ; i++ )
         {
             if( strcasecmp( fmt.slds2.recfm, valfm[ i ].recfm ) == 0 )
             {
@@ -192,7 +193,7 @@ merge( SLLABEL *lab )
     /*
     || Locate final RECFM string
     */
-    for( i = 0 ; i < VALFMCNT ; i++ )
+    for( i = 0 ; i < (int)VALFMCNT ; i++ )
     {
         if( strcasecmp( fmt.slds2.recfm, valfm[ i ].recfm ) == 0 )
         {
@@ -618,9 +619,6 @@ main( int argc, char *argv[] )
     int rc;
     int i;
 
-    /* Display the program identification message */
-    display_version (stderr, "Hercules HET extract files program ");
-
 #ifdef EXTERNALGUI
     if (argc >= 1 && strncmp(argv[argc-1],"EXTERNALGUI",11) == 0)
     {
@@ -629,6 +627,9 @@ main( int argc, char *argv[] )
     }
 #endif /*EXTERNALGUI*/
 
+    /* Display the program identification message */
+    display_version (stderr, "Hercules HET extract files program ");
+
     /*
     || Process option switches
     */
@@ -723,7 +724,7 @@ main( int argc, char *argv[] )
         || Lookup the specified RECFM in our table
         */
         opts.recfm = 0;
-        for( i = 0 ; i < VALFMCNT ; i++ )
+        for( i = 0 ; i < (int)VALFMCNT ; i++ )
         {
             if( strcasecmp( argv[ optind + 3 ], valfm[ i ].recfm ) == 0 )
             {
@@ -741,7 +742,7 @@ main( int argc, char *argv[] )
             || Dump out the valid RECFMs
             */
             printf( "Valid record formats are:\n" );
-            for( i = 0 ; i < VALFMCNT ; i++ )
+            for( i = 0 ; i < (int)VALFMCNT ; i++ )
             {
                 printf( "  %-4.4s", valfm[ i ].recfm );
                 if( ( ( i + 1 ) % 3 ) == 0 )
