@@ -1,7 +1,7 @@
 /*
 || ----------------------------------------------------------------------------
 ||
-|| HETMAP.C     (c) Copyright Leland Lucius, 2000-2001
+|| HETMAP.C     (c) Copyright Leland Lucius, 2000-2003
 ||              Released under terms of the Q Public License.
 ||
 || Displays information about the structure of a Hercules Emulated Tape. 
@@ -38,7 +38,7 @@ static const char help[] =
 #ifdef EXTERNALGUI
 /* Special flag to indicate whether or not we're being
    run under the control of the external GUI facility. */
-static int extgui = 0;
+int extgui = 0;
 /* Previous reported file position */
 static long prevpos = 0;
 /* Report progress every this many bytes */
@@ -144,28 +144,23 @@ main( int argc, char *argv[] )
     HETB *hetb;
     int rc;
     int fileno;
-    int blocks;
-    int uminsz;
-    int umaxsz;
-    int ubytes;
-    int cminsz;
-    int cmaxsz;
-    int cbytes;
-    int totblocks;
-    int totubytes;
-    int totcbytes;
-    int opts;
+    U32  blocks;
+    U32  uminsz;
+    U32  umaxsz;
+    U32  ubytes;
+    U32  cminsz;
+    U32  cmaxsz;
+    U32  cbytes;
+    U32  totblocks;
+    U32  totubytes;
+    U32  totcbytes;
+    U32  opts;
 
 #define O_ALL           0xC0
 #define O_FILES         0X80
 #define O_LABELS        0X40
 #define O_DATASETS      0X20
 
-    opts = O_ALL;
-
-    /* Display the program identification message */
-    display_version (stderr, "Hercules HET map program ");
-
 #ifdef EXTERNALGUI
     if (argc >= 1 && strncmp(argv[argc-1],"EXTERNALGUI",11) == 0)
     {
@@ -174,6 +169,11 @@ main( int argc, char *argv[] )
     }
 #endif /*EXTERNALGUI*/
 
+    opts = O_ALL;
+
+    /* Display the program identification message */
+    display_version (stderr, "Hercules HET map program ");
+
     while( TRUE )
     {
         rc = getopt( argc, argv, "adfhlt" );
