@@ -1,8 +1,8 @@
-/* PLO.C        (c) Copyright Jan Jaeger, 2000-2001                  */
+/* PLO.C        (c) Copyright Jan Jaeger, 2000-2003                  */
 /*              Perform Locked Operation functions codes             */
 
-/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2001      */
-/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2001      */
+/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2003      */
+/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2003      */
 
 #include "hercules.h"
 
@@ -51,6 +51,8 @@ U64 op1c,
 U32 op4alet = 0;
 VADR op4addr;
 
+    UNREFERENCED(r1);
+
     DW_CHECK(effective_addr4, regs);
     DW_CHECK(effective_addr2, regs);
 
@@ -142,6 +144,8 @@ BYTE op1c[16],
 U32 op4alet = 0;
 VADR op4addr;
 
+    UNREFERENCED(r1);
+
     DW_CHECK(effective_addr4, regs);
     DW_CHECK(effective_addr2, regs);
 
@@ -193,6 +197,10 @@ int ARCH_DEP(plo_cs) (int r1, int r3, VADR effective_addr2, int b2,
 {
 U32 op2;
 
+    UNREFERENCED(r3);
+    UNREFERENCED(effective_addr4);
+    UNREFERENCED(b4);
+
     ODD_CHECK(r1, regs);
     FW_CHECK(effective_addr2, regs);
 
@@ -225,6 +233,9 @@ U64 op1c,
     op1r,
     op2;
 
+    UNREFERENCED(r1);
+    UNREFERENCED(r3);
+
     DW_CHECK(effective_addr4, regs);
     DW_CHECK(effective_addr2, regs);
 
@@ -260,6 +271,10 @@ int ARCH_DEP(plo_csgr) (int r1, int r3, VADR effective_addr2, int b2,
 {
 U64 op2;
 
+    UNREFERENCED(r3);
+    UNREFERENCED(effective_addr4);
+    UNREFERENCED(b4);
+
     ODD_CHECK(r1, regs);
     DW_CHECK(effective_addr2, regs);
 
@@ -294,6 +309,9 @@ BYTE op1c[16],
      op1r[16],
      op2[16];
 
+    UNREFERENCED(r1);
+    UNREFERENCED(r3);
+
     DW_CHECK(effective_addr4, regs);
     DW_CHECK(effective_addr2, regs);
 
@@ -384,6 +402,8 @@ U64 op1c,
 U32 op4alet = 0;
 VADR op4addr;
 
+    UNREFERENCED(r1);
+
     DW_CHECK(effective_addr2, regs);
     DW_CHECK(effective_addr4, regs);
 
@@ -520,6 +540,8 @@ BYTE op1c[16],
 U32 op4alet = 0;
 VADR op4addr;
 
+    UNREFERENCED(r1);
+
     DW_CHECK(effective_addr2, regs);
     DW_CHECK(effective_addr4, regs);
 
@@ -635,6 +657,8 @@ U64 op1c,
 U32 op4alet = 0;
 VADR op4addr;
 
+    UNREFERENCED(r1);
+
     DW_CHECK(effective_addr2, regs);
     DW_CHECK(effective_addr4, regs);
 
@@ -734,6 +758,8 @@ BYTE op1c[16],
 U32 op4alet = 0;
 VADR op4addr;
 
+    UNREFERENCED(r1);
+
     DW_CHECK(effective_addr2, regs);
     DW_CHECK(effective_addr4, regs);
 
@@ -884,6 +910,8 @@ U32 op4alet = 0,
 VADR op4addr,
     op6addr;
 
+    UNREFERENCED(r1);
+
     DW_CHECK(effective_addr2, regs);
     DW_CHECK(effective_addr4, regs);
 
@@ -1060,6 +1088,8 @@ U32 op4alet = 0,
 VADR op4addr,
     op6addr;
 
+    UNREFERENCED(r1);
+
     DW_CHECK(effective_addr2, regs);
     DW_CHECK(effective_addr4, regs);
 
@@ -1266,6 +1296,8 @@ VADR op4addr,
     op6addr,
     op8addr;
 
+    UNREFERENCED(r1);
+
     DW_CHECK(effective_addr2, regs);
     DW_CHECK(effective_addr4, regs);
 
@@ -1498,6 +1530,8 @@ VADR op4addr,
     op6addr,
     op8addr;
 
+    UNREFERENCED(r1);
+
     DW_CHECK(effective_addr2, regs);
     DW_CHECK(effective_addr4, regs);
 
@@ -1599,11 +1633,15 @@ VADR op4addr,
 
 #if !defined(_GEN_ARCH)
 
-#define  _GEN_ARCH 390
-#include "plo.c"
+#if defined(_ARCHMODE2)
+ #define  _GEN_ARCH _ARCHMODE2
+ #include "plo.c"
+#endif
 
+#if defined(_ARCHMODE3)
 #undef   _GEN_ARCH
-#define  _GEN_ARCH 370
+#define  _GEN_ARCH _ARCHMODE3
 #include "plo.c"
+#endif
 
 #endif /*!defined(_GEN_ARCH)*/
