@@ -1,8 +1,8 @@
-/* VSTORE.H     (c) Copyright Roger Bowler, 1999-2001                */
+/* VSTORE.H     (c) Copyright Roger Bowler, 1999-2003                */
 /*              ESA/390 Dynamic Address Translation                  */
 
-/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2001      */
-/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2001      */
+/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2003      */
+/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2003      */
 
 /*-------------------------------------------------------------------*/
 /* This module implements the DAT, ALET, and ASN translation         */
@@ -55,17 +55,17 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
        depending on whether operand crosses a page boundary */
     if (addr2 == (addr & PAGEFRAME_PAGEMASK)) {
         addr = LOGICAL_TO_ABS (addr, arn, regs, ACCTYPE_WRITE, akey);
-        memcpy (sysblk.mainstor+addr, src, len+1);
+        memcpy (regs->mainstor+addr, src, len+1);
     } else {
         len1 = addr2 - addr;
         len2 = len - len1 + 1;
         addr = LOGICAL_TO_ABS_SKP (addr, arn, regs, ACCTYPE_WRITE_SKP, akey);
         addr2 = LOGICAL_TO_ABS_SKP (addr2, arn, regs, ACCTYPE_WRITE_SKP, akey);
         /* both pages are accessable, so set ref & change bits */
-        STORAGE_KEY(addr) |= (STORKEY_REF | STORKEY_CHANGE);
-        STORAGE_KEY(addr2) |= (STORKEY_REF | STORKEY_CHANGE);
-        memcpy (sysblk.mainstor+addr, src, len1);
-        memcpy (sysblk.mainstor+addr2, src+len1, len2);
+        STORAGE_KEY(addr, regs) |= (STORKEY_REF | STORKEY_CHANGE);
+        STORAGE_KEY(addr2, regs) |= (STORKEY_REF | STORKEY_CHANGE);
+        memcpy (regs->mainstor+addr, src, len1);
+        memcpy (regs->mainstor+addr2, src+len1, len2);
     }
 } /* end function ARCH_DEP(vstorec) */
 
@@ -87,7 +87,7 @@ _VSTORE_C_STATIC void ARCH_DEP(vstoreb) (BYTE value, VADR addr,
 {
     addr = LOGICAL_TO_ABS (addr, arn, regs, ACCTYPE_WRITE,
                                 regs->psw.pkey);
-    sysblk.mainstor[addr] = value;
+    regs->mainstor[addr] = value;
 } /* end function ARCH_DEP(vstoreb) */
 
 /*-------------------------------------------------------------------*/
@@ -116,10 +116,10 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
     abs1 = LOGICAL_TO_ABS_SKP (addr, arn, regs, ACCTYPE_WRITE_SKP, regs->psw.pkey);
 
     /* Check if store crosses page or not */
-    if (!(addr & 1) || (addr & PAGEFRAME_BYTEMASK) <= (PAGEFRAME_PAGESIZE - 2))
+    if ((addr & PAGEFRAME_BYTEMASK) <= (PAGEFRAME_PAGESIZE - 2))
     {
-        STORAGE_KEY(abs1) |= (STORKEY_REF | STORKEY_CHANGE);
-        STORE_HW(sysblk.mainstor + abs1, value);
+        STORAGE_KEY(abs1, regs) |= (STORKEY_REF | STORKEY_CHANGE);
+        STORE_HW(regs->mainstor + abs1, value);
         return;
     }
 
@@ -128,12 +128,12 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
     abs2 = LOGICAL_TO_ABS_SKP (addr2, arn, regs, ACCTYPE_WRITE_SKP, regs->psw.pkey);
 
     /* Both pages are accessable, now safe to set Reference and Change bits */
-    STORAGE_KEY(abs1) |= (STORKEY_REF | STORKEY_CHANGE);
-    STORAGE_KEY(abs2) |= (STORKEY_REF | STORKEY_CHANGE);
+    STORAGE_KEY(abs1, regs) |= (STORKEY_REF | STORKEY_CHANGE);
+    STORAGE_KEY(abs2, regs) |= (STORKEY_REF | STORKEY_CHANGE);
 
     /* Store integer value at operand location */
-    sysblk.mainstor[abs1] = value >> 8;
-    sysblk.mainstor[abs2] = value & 0xFF;
+    regs->mainstor[abs1] = value >> 8;
+    regs->mainstor[abs2] = value & 0xFF;
 
 } /* end function ARCH_DEP(vstore2) */
 
@@ -165,10 +165,10 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
     abs = LOGICAL_TO_ABS_SKP (addr, arn, regs, ACCTYPE_WRITE_SKP, regs->psw.pkey);
 
     /* Check if store crosses page or not */
-    if (!(addr & 3) || (addr & PAGEFRAME_BYTEMASK) <= (PAGEFRAME_PAGESIZE - 4))
+    if ((addr & PAGEFRAME_BYTEMASK) <= (PAGEFRAME_PAGESIZE - 4))
     {
-        STORAGE_KEY(abs) |= (STORKEY_REF | STORKEY_CHANGE);
-        STORE_FW(sysblk.mainstor + abs, value);
+        STORAGE_KEY(abs, regs) |= (STORKEY_REF | STORKEY_CHANGE);
+        STORE_FW(regs->mainstor + abs, value);
         return;
     }
 
@@ -177,14 +177,14 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
     abs2 = LOGICAL_TO_ABS_SKP (addr2, arn, regs, ACCTYPE_WRITE_SKP, regs->psw.pkey);
 
     /* Both pages are accessable, now safe to set Reference and Change bits */
-    STORAGE_KEY(abs) |= (STORKEY_REF | STORKEY_CHANGE);
-    STORAGE_KEY(abs2) |= (STORKEY_REF | STORKEY_CHANGE);
+    STORAGE_KEY(abs, regs) |= (STORKEY_REF | STORKEY_CHANGE);
+    STORAGE_KEY(abs2, regs) |= (STORKEY_REF | STORKEY_CHANGE);
 
     /* Store integer value byte by byte at operand location */
     for (i=0, k=24; i < 4; i++, k -= 8) {
 
         /* Store byte in absolute storage */
-        sysblk.mainstor[abs] = (value >> k) & 0xFF;
+        regs->mainstor[abs] = (value >> k) & 0xFF;
 
         /* Increment absolute address and virtual address */
         abs++;
@@ -226,10 +226,10 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
 
     abs = LOGICAL_TO_ABS_SKP (addr, arn, regs, ACCTYPE_WRITE_SKP, regs->psw.pkey);
     /* Check if store crosses page or not */
-    if (!(addr & 7) || (addr & PAGEFRAME_BYTEMASK) <= (PAGEFRAME_PAGESIZE - 8))
+    if ((addr & PAGEFRAME_BYTEMASK) <= (PAGEFRAME_PAGESIZE - 8))
     {
-        STORAGE_KEY(abs) |= (STORKEY_REF | STORKEY_CHANGE);
-        STORE_DW(sysblk.mainstor + abs, value);
+        STORAGE_KEY(abs, regs) |= (STORKEY_REF | STORKEY_CHANGE);
+        STORE_DW(regs->mainstor + abs, value);
         return;
     }
 
@@ -238,14 +238,14 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
     abs2 = LOGICAL_TO_ABS_SKP (addr2, arn, regs, ACCTYPE_WRITE_SKP, regs->psw.pkey);
 
     /* Both pages are accessable, now safe to set Reference and Change bits */
-    STORAGE_KEY(abs) |= (STORKEY_REF | STORKEY_CHANGE);
-    STORAGE_KEY(abs2) |= (STORKEY_REF | STORKEY_CHANGE);
+    STORAGE_KEY(abs, regs) |= (STORKEY_REF | STORKEY_CHANGE);
+    STORAGE_KEY(abs2, regs) |= (STORKEY_REF | STORKEY_CHANGE);
 
     /* Store integer value byte by byte at operand location */
     for (i=0, k=56; i < 8; i++, k -= 8) {
 
         /* Store byte in absolute storage */
-        sysblk.mainstor[abs] = (value >> k) & 0xFF;
+        regs->mainstor[abs] = (value >> k) & 0xFF;
 
         /* Increment absolute address and virtual address */
         abs++;
@@ -295,14 +295,14 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
        (Page boundary set at 800 to catch FPO crosser too) */
     if (addr2 == (addr & ~0x7FF)) {
         addr = LOGICAL_TO_ABS (addr, arn, regs, ACCTYPE_READ, akey);
-        memcpy (dest, sysblk.mainstor+addr, len+1);
+        memcpy (dest, regs->mainstor+addr, len+1);
     } else {
         len1 = addr2 - addr;
         len2 = len - len1 + 1;
         addr = LOGICAL_TO_ABS (addr, arn, regs, ACCTYPE_READ, akey);
         addr2 = LOGICAL_TO_ABS (addr2, arn, regs, ACCTYPE_READ, akey);
-        memcpy (dest, sysblk.mainstor+addr, len1);
-        memcpy (dest+len1, sysblk.mainstor+addr2, len2);
+        memcpy (dest, regs->mainstor+addr, len1);
+        memcpy (dest+len1, regs->mainstor+addr2, len2);
     }
 } /* end function ARCH_DEP(vfetchc) */
 
@@ -325,7 +325,7 @@ _VSTORE_C_STATIC BYTE ARCH_DEP(vfetchb) (VADR addr, int arn,
 {
     addr = LOGICAL_TO_ABS (addr, arn, regs, ACCTYPE_READ,
                                 regs->psw.pkey);
-    return sysblk.mainstor[addr];
+    return regs->mainstor[addr];
 } /* end function ARCH_DEP(vfetchb) */
 
 /*-------------------------------------------------------------------*/
@@ -356,8 +356,8 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
 
     /* Fetch 2 bytes when operand does not cross page boundary
        (Page boundary test at 800 to catch FPO crosser too) */
-    if(!(addr & 1) || (abs1 & 0x000007FF) <= (2048 - 2))
-        return fetch_hw(sysblk.mainstor + abs1);
+    if((abs1 & 0x000007FF) <= (2048 - 2))
+        return fetch_hw(regs->mainstor + abs1);
 
     /* Calculate address of second byte of operand */
     addr2 = (addr + 1) & ADDRESS_MAXWRAP(regs);
@@ -366,7 +366,7 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
     abs2 = LOGICAL_TO_ABS (addr2, arn, regs, ACCTYPE_READ, akey);
 
     /* Return integer value of operand */
-    return (sysblk.mainstor[abs1] << 8) | sysblk.mainstor[abs2];
+    return (regs->mainstor[abs1] << 8) | regs->mainstor[abs2];
 } /* end function ARCH_DEP(vfetch2) */
 
 /*-------------------------------------------------------------------*/
@@ -399,8 +399,8 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
 
     /* Fetch 4 bytes when operand does not cross page boundary
        (Page boundary test at 800 to catch FPO crosser too) */
-    if(!(addr & 3) || (abs & 0x000007FF) <= (2048 - 4))
-        return fetch_fw(sysblk.mainstor + abs);
+    if((abs & 0x000007FF) <= (2048 - 4))
+        return fetch_fw(regs->mainstor + abs);
 
     /* Operand is not fullword aligned and may cross a page boundary */
 
@@ -414,7 +414,7 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
 
         /* Fetch byte from absolute storage */
         value <<= 8;
-        value |= sysblk.mainstor[abs];
+        value |= regs->mainstor[abs];
 
         /* Increment absolute address and virtual address */
         abs++;
@@ -460,8 +460,8 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
 
     /* Fetch 8 bytes when operand does not cross page boundary
        (Page boundary test at 800 to catch FPO crosser too) */
-    if(!(addr & 7) || (abs & 0x000007FF) <= (2048 - 8))
-        return fetch_dw(sysblk.mainstor + abs);
+    if((abs & 0x000007FF) <= (2048 - 8))
+        return fetch_dw(regs->mainstor + abs);
 
     /* Calculate page address of last byte of operand */
     addr2 = (addr + 7) & ADDRESS_MAXWRAP(regs);
@@ -473,7 +473,7 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
 
         /* Fetch byte from absolute storage */
         value <<= 8;
-        value |= sysblk.mainstor[abs];
+        value |= regs->mainstor[abs];
 
         /* Increment absolute address and virtual address */
         abs++;
@@ -515,32 +515,80 @@ _VFETCH_C_STATIC void ARCH_DEP(instfetch) (BYTE *dest, VADR addr,
 RADR    abs;                            /* Absolute storage address  */
 BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
 
-    /* Obtain current access key from PSW */
-    akey = regs->psw.pkey;
 
     /* Program check if instruction address is odd */
     if (addr & 0x01)
         ARCH_DEP(program_interrupt) (regs, PGM_SPECIFICATION_EXCEPTION);
 
+
+#if defined(FEATURE_PER)
+    /* Save the address address used to fetch the instruction */
+    if( EN_IC_PER(regs) )
+    {
+#if defined(FEATURE_PER2)
+        regs->perc = 0x40    /* ATMID-validity */
+                   | (regs->psw.amode64 << 7)
+                   | (regs->psw.amode << 5)
+                   | (!REAL_MODE(&regs->psw) ? 0x10 : 0)
+                   | (regs->psw.space << 3)
+                   | (regs->psw.armode << 2);
+#else /*!defined(FEATURE_PER2)*/
+        regs->perc = 0;
+#endif /*!defined(FEATURE_PER2)*/
+
+        /* For EXecute instvalid will be true */
+        if(!regs->instvalid)
+            regs->peradr = addr;
+    }
+
+
+    if( EN_IC_PER_IF(regs)
+      && PER_RANGE_CHECK(addr,regs->CR(10),regs->CR(11)) )
+        ON_IC_PER_IF(regs);
+#endif /*defined(FEATURE_PER)*/
+
+    /* Obtain current access key from PSW */
+    akey = regs->psw.pkey;
+
+
+
     /* Fetch six bytes if instruction cannot cross a page boundary */
     if ((addr & 0x7FF) <= (0x800 - 6))
     {
         abs = LOGICAL_TO_ABS (addr, 0, regs, ACCTYPE_INSTFETCH, akey);
 #if defined(OPTION_AIA_BUFFER)
-        regs->AI = abs & PAGEFRAME_PAGEMASK;
-        regs->VI = addr & PAGEFRAME_PAGEMASK;
+#if defined(FEATURE_PER)
+        if( !EN_IC_PER(regs) )
+#endif /*defined(FEATURE_PER)*/
+        {
+            regs->AI = abs & PAGEFRAME_PAGEMASK;
+            regs->VI = addr & PAGEFRAME_PAGEMASK;
+        }
+#if defined(FEATURE_PER)
+        else
+            INVALIDATE_AIA(regs);
+#endif /*defined(FEATURE_PER)*/
 #endif /*defined(OPTION_AIA_BUFFER)*/
-        memcpy (dest, sysblk.mainstor+abs, 6);
+        memcpy (dest, regs->mainstor+abs, 6);
         return;
     }
 
     /* Fetch first two bytes of instruction */
     abs = LOGICAL_TO_ABS (addr, 0, regs, ACCTYPE_INSTFETCH, akey);
 #if defined(OPTION_AIA_BUFFER)
-    regs->AI = abs & PAGEFRAME_PAGEMASK;
-    regs->VI = addr & PAGEFRAME_PAGEMASK;
+#if defined(FEATURE_PER)
+    if( !EN_IC_PER(regs) )
+#endif /*defined(FEATURE_PER)*/
+    {
+        regs->AI = abs & PAGEFRAME_PAGEMASK;
+        regs->VI = addr & PAGEFRAME_PAGEMASK;
+    }
+#if defined(FEATURE_PER)
+        else
+            INVALIDATE_AIA(regs);
+#endif /*defined(FEATURE_PER)*/
 #endif /*defined(OPTION_AIA_BUFFER)*/
-    memcpy (dest, sysblk.mainstor+abs, 2);
+    memcpy (dest, regs->mainstor+abs, 2);
 
     /* Return if two-byte instruction */
     if (dest[0] < 0x40) return;
@@ -552,7 +600,7 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
     if ((addr & 0x7FF) == 0x000) {
         abs = LOGICAL_TO_ABS (addr, 0, regs, ACCTYPE_INSTFETCH, akey);
     }
-    memcpy (dest+2, sysblk.mainstor+abs, 2);
+    memcpy (dest+2, regs->mainstor+abs, 2);
 
     /* Return if four-byte instruction */
     if (dest[0] < 0xC0) return;
@@ -564,7 +612,7 @@ BYTE    akey;                           /* Bits 0-3=key, 4-7=zeroes  */
     if ((addr & 0x7FF) == 0x000) {
         abs = LOGICAL_TO_ABS (addr, 0, regs, ACCTYPE_INSTFETCH, akey);
     }
-    memcpy (dest+4, sysblk.mainstor+abs, 2);
+    memcpy (dest+4, regs->mainstor+abs, 2);
 
 } /* end function ARCH_DEP(instfetch) */
 #endif
@@ -616,6 +664,7 @@ int     i;                              /* Loop counter              */
 BYTE    obyte;                          /* Operand byte              */
 #ifdef OPTION_FAST_MOVECHAR
 BYTE    slow = 0;
+BYTE    *a1, *a2;
 #endif
 
     /* Translate addresses of leftmost operand bytes */
@@ -649,9 +698,9 @@ BYTE    slow = 0;
     }
 
     /* all operands and page crossers valid, now alter ref & chg bits */
-    STORAGE_KEY(abs1) |= (STORKEY_REF | STORKEY_CHANGE);
+    STORAGE_KEY(abs1, regs) |= (STORKEY_REF | STORKEY_CHANGE);
     if (npa1)
-        STORAGE_KEY(npa1) |= (STORKEY_REF | STORKEY_CHANGE);
+        STORAGE_KEY(npa1, regs) |= (STORKEY_REF | STORKEY_CHANGE);
 
 #ifdef FEATURE_INTERVAL_TIMER
     /* Special case for mvc to/from interval timer */
@@ -660,7 +709,7 @@ BYTE    slow = 0;
         /* We've got a 4-byte wide, 4-byte aligned access of the interval timer */
         obtain_lock( &sysblk.todlock );
 
-        *(U32 *)&sysblk.mainstor[abs1] = *(U32 *)&sysblk.mainstor[abs2];
+        *(U32 *)&regs->mainstor[abs1] = *(U32 *)&regs->mainstor[abs2];
 
         release_lock( &sysblk.todlock );
     } else {
@@ -669,8 +718,14 @@ BYTE    slow = 0;
 #ifdef OPTION_FAST_MOVECHAR
     if (!slow)
     {
-        for (i = 0; i < len + 1; i++)
-            sysblk.mainstor[abs1++] = sysblk.mainstor[abs2++];
+    /* Gabor Hoffer (performance option) */
+        a1 = regs->mainstor+abs1;
+    a2 = regs->mainstor+abs2;
+    for (i = 0; i < len + 1; i++) a1 [i] = a2 [i];
+/*  
+    for (i = 0; i < len + 1; i++)
+            regs->mainstor[abs1++] = regs->mainstor[abs2++];
+*/
         return;
     }
 #endif
@@ -679,10 +734,10 @@ BYTE    slow = 0;
     for ( i = 0; i < len+1; i++ )
     {
         /* Fetch a byte from the source operand */
-        obyte = sysblk.mainstor[abs2];
+        obyte = regs->mainstor[abs2];
 
         /* Store the byte in the destination operand */
-        sysblk.mainstor[abs1] = obyte;
+        regs->mainstor[abs1] = obyte;
 
         /* Increment first operand address */
         addr1++;
