@@ -1,7 +1,7 @@
 /* XSTORE.C   Expanded storage related instructions - Jan Jaeger     */
 
-/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2001      */
-/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2001      */
+/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2003      */
+/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2003      */
 
 /* MVPG moved from cpu.c to xstore.c   05/07/00 Jan Jaeger */
 
@@ -68,7 +68,7 @@ U32     xaddr;                          /* Expanded storage address  */
     maddr &= XSTORE_PAGEMASK;
 
     /* Copy data from expanded to main */
-    memcpy (sysblk.mainstor + maddr, sysblk.xpndstor + xaddr,
+    memcpy (regs->mainstor + maddr, sysblk.xpndstor + xaddr,
             XSTORE_PAGESIZE);
 
     /* cc0 means pgin ok */
@@ -131,11 +131,11 @@ U32     xaddr;                          /* Expanded storage address  */
 
     /* Obtain abs address, verify access and set ref/change bits */
     maddr = LOGICAL_TO_ABS (regs->GR(r1) & ADDRESS_MAXWRAP(regs),
-         USE_REAL_ADDR, regs, ACCTYPE_READ, regs->psw.pkey);
+         USE_REAL_ADDR, regs, ACCTYPE_READ, 0);
     maddr &= XSTORE_PAGEMASK;
 
     /* Copy data from main to expanded */
-    memcpy (sysblk.xpndstor + xaddr, sysblk.mainstor + maddr,
+    memcpy (sysblk.xpndstor + xaddr, regs->mainstor + maddr,
             XSTORE_PAGESIZE);
 
     /* cc0 means pgout ok */
@@ -157,6 +157,11 @@ int     r1, r2;                         /* Values of R fields        */
 
     PRIV_CHECK(regs);
 
+#if defined(_FEATURE_SIE)
+    if(regs->sie_state && !(regs->siebk->ec[0] & SIE_EC0_MVPG))
+        longjmp(regs->progjmp, SIE_INTERCEPT_INST);
+#endif /*defined(_FEATURE_SIE)*/
+
     /* Perform serialization before operation */
     PERFORM_SERIALIZATION (regs);
 
@@ -197,9 +202,9 @@ DEF_INST(move_page)
 {
 int     r1, r2;                         /* Register values           */
 int     rc = 0;                         /* Return code               */
-int     cc = 0;				/* Condition code            */
+int     cc = 0;             /* Condition code            */
 VADR    vaddr1, vaddr2;                 /* Virtual addresses         */
-RADR    raddr1, raddr2;                 /* Real addresses            */
+RADR    raddr1, raddr2, xpkeya;         /* Real addresses            */
 RADR    aaddr1 = 0, aaddr2 = 0;         /* Absolute addresses        */
 int     priv;                           /* 1=Private address space   */
 int     prot = 0;                       /* 1=Protected page          */
@@ -216,11 +221,10 @@ BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
 
     RRE(inst, execflag, regs, r1, r2);
 
-    /* Specification exception if register 0 bits 16-19 are
-       not all zero, or if bits 20 and 21 are both ones */
-    if ((regs->GR_L(0) & 0x0000F000) != 0
-        || (regs->GR_L(0) & 0x00000C00) == 0x00000C00)
-        ARCH_DEP(program_interrupt) (regs, PGM_SPECIFICATION_EXCEPTION);
+#if defined(_FEATURE_SIE)
+    if(regs->sie_state && !(regs->siebk->ec[0] & SIE_EC0_MVPG))
+        longjmp(regs->progjmp, SIE_INTERCEPT_INST);
+#endif /*defined(_FEATURE_SIE)*/
 
     /* Use PSW key as access key for both operands */
     akey1 = akey2 = regs->psw.pkey;
@@ -246,6 +250,12 @@ BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
             akey2 = akey;
     }
 
+    /* Specification exception if register 0 bits 16-19 are
+       not all zero, or if bits 20 and 21 are both ones */
+    if ((regs->GR_L(0) & 0x0000F000) != 0
+        || (regs->GR_L(0) & 0x00000C00) == 0x00000C00)
+        ARCH_DEP(program_interrupt) (regs, PGM_SPECIFICATION_EXCEPTION);
+
     /* Determine the logical addresses of each operand */
     vaddr1 = regs->GR(r1) & ADDRESS_MAXWRAP(regs);
     vaddr2 = regs->GR(r2) & ADDRESS_MAXWRAP(regs);
@@ -255,26 +265,60 @@ BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
     vaddr2 &= XSTORE_PAGEMASK;
 
     /* Obtain the real or expanded address of each operand */
-    if (!REAL_MODE(&regs->psw))
+    if (!REAL_MODE(&regs->psw)
+#if defined(_FEATURE_SIE)
+        || regs->sie_state
+#endif /*defined(_FEATURE_SIE)*/
+                              )
     {
         /* Translate the second operand address to a real address */
-        rc = ARCH_DEP(translate_addr) (vaddr2, r2, regs, ACCTYPE_READ, &raddr2,
-                        &xcode, &priv, &prot, &stid);
+        if(!REAL_MODE(&regs->psw))
+            rc = ARCH_DEP(translate_addr) (vaddr2, r2, regs, ACCTYPE_READ,
+                        &raddr2, &xcode, &priv, &prot, &stid);
+        else
+            raddr2 = vaddr2;
 
-	raddr2 = APPLY_PREFIXING (raddr2, regs->PX);
+    raddr2 = APPLY_PREFIXING (raddr2, regs->PX);
 
-        if (raddr2 < regs->mainsize)
-            SIE_TRANSLATE(&raddr2,ACCTYPE_READ,regs);
+        if (raddr2 > regs->mainlim)
+            ARCH_DEP(program_interrupt) (regs, PGM_ADDRESSING_EXCEPTION);
+
+#if defined(_FEATURE_SIE)
+        if(regs->sie_state  && !regs->sie_pref)
+        {
+        U32 sie_stid;
+        U16 sie_xcode;
+        int sie_private;
+
+#if defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE)
+            if (SIE_TRANSLATE_ADDR (regs->sie_mso + raddr2,
+                ((regs->siebk->mx & SIE_MX_XC) && regs->psw.armode && r2 > 0) ?
+                    r2 :
+                    USE_PRIMARY_SPACE,
+                    regs->hostregs, ACCTYPE_SIE, &raddr2, &sie_xcode,
+                    &sie_private, &prot, &sie_stid))
+#else /*!defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE)*/
+            if (SIE_TRANSLATE_ADDR (regs->sie_mso + raddr2,
+                    USE_PRIMARY_SPACE,
+                    regs->hostregs, ACCTYPE_SIE, &raddr2, &sie_xcode,
+                    &sie_private, &prot, &sie_stid))
+#endif /*!defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE)*/
+                (regs->sie_hostpi) (regs->hostregs, sie_xcode);
+
+            /* Convert host real address to host absolute address */
+            raddr2 = APPLY_PREFIXING (raddr2, regs->hostregs->PX);
+        }
+#endif /*defined(_FEATURE_SIE)*/
 
 #if defined(FEATURE_EXPANDED_STORAGE)
         if(rc == 2)
         {
-	    FETCH_W(pte2,sysblk.mainstor + raddr2);
+        FETCH_W(pte2,regs->mainstor + raddr2);
             /* If page is invalid in real storage but valid in expanded
                storage then xpblk2 now contains expanded storage block# */
             if(pte2 & PAGETAB_ESVALID)
             {
-                xpblk2 = pte2 & ZPGETAB_PFRA >> 12;
+                xpblk2 = (pte2 & ZPGETAB_PFRA) >> 12;
 #if defined(_FEATURE_SIE)
                 if(regs->sie_state)
                 {
@@ -290,27 +334,24 @@ BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
                 }
 #endif /*defined(_FEATURE_SIE)*/
 
-                /* If the expanded storage block does not exist
-                   then terminate the instruction */
-                if (xpblk2 >= sysblk.xpndsize)
-                {
-                    cc = 2;
-                    goto mvpg_progck;
-                }
-
                 rc = 0;
                 xpvalid2 = 1;
-                xpkey2 = sysblk.mainstor[raddr2 +
+                xpkeya = raddr2 +
 #if defined(FEATURE_ESAME)
-                                                  2048
+                                   2048;
 #else /*!defined(FEATURE_ESAME)*/
                 /* For ESA/390 mode, the XPTE lies directly beyond 
                    the PTE, and each entry is 12 bytes long, we must
                    therefor add 1024 + 8 times the page index */
-	                             1024 + ((vaddr1 & 0x000FF000) >> 9)
+                                 1024 + ((vaddr2 & 0x000FF000) >> 9);
 #endif /*!defined(FEATURE_ESAME)*/
-                                                      ];
-/*DEBUG*/ logmsg("MVPG pte2 = " F_CREG ", xkey2 = %2.2X\n",pte2,xpkey2);
+                if (xpkeya > regs->mainlim)
+                    ARCH_DEP(program_interrupt) (regs, PGM_ADDRESSING_EXCEPTION);
+                xpkey2 = regs->mainstor[xpkeya]; 
+
+/*DEBUG logmsg("MVPG pte2 = " F_CREG ", xkey2 = %2.2X, xpblk2 = %5.5X, akey2 = %2.2X\n",
+                  pte2,xpkey2,xpblk2,akey2);  */
+
             }
             else
             {
@@ -324,30 +365,60 @@ BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
            in either main storage or expanded storage */
         if (rc)
         {
-	    cc = 2;
+        cc = 2;
             goto mvpg_progck;
-	}
+    }
 
         /* Reset protection indication before calling translate_addr() */
         prot = 0;
         /* Translate the first operand address to a real address */
-        rc = ARCH_DEP(translate_addr) (vaddr1, r1, regs, ACCTYPE_WRITE, &raddr1,
-                        &xcode, &priv, &prot, &stid);
+        if(!REAL_MODE(&regs->psw))
+            rc = ARCH_DEP(translate_addr) (vaddr1, r1, regs, ACCTYPE_WRITE,
+                        &raddr1, &xcode, &priv, &prot, &stid);
+        else
+            raddr1 = vaddr1;
 
-	raddr1 = APPLY_PREFIXING (raddr1, regs->PX);
+    raddr1 = APPLY_PREFIXING (raddr1, regs->PX);
 
-        if (raddr1 < regs->mainsize)
-            SIE_TRANSLATE(&raddr1,ACCTYPE_READ,regs);
+        if (raddr1 > regs->mainlim)
+            ARCH_DEP(program_interrupt) (regs, PGM_ADDRESSING_EXCEPTION);
+
+#if defined(_FEATURE_SIE)
+        if(regs->sie_state  && !regs->sie_pref)
+        {
+        U32 sie_stid;
+        U16 sie_xcode;
+        int sie_private;
+
+#if defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE)
+            if (SIE_TRANSLATE_ADDR (regs->sie_mso + raddr1,
+                ((regs->siebk->mx & SIE_MX_XC) && regs->psw.armode && r1 > 0) ?
+                    r1 :
+                    USE_PRIMARY_SPACE,
+                    regs->hostregs, ACCTYPE_SIE, &raddr1, &sie_xcode,
+                    &sie_private, &prot, &sie_stid))
+#else /*!defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE)*/
+            if (SIE_TRANSLATE_ADDR (regs->sie_mso + raddr1,
+                    USE_PRIMARY_SPACE,
+                    regs->hostregs, ACCTYPE_SIE, &raddr1, &sie_xcode,
+                    &sie_private, &prot, &sie_stid))
+#endif /*!defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE)*/
+                (regs->sie_hostpi) (regs->hostregs, sie_xcode);
+
+            /* Convert host real address to host absolute address */
+            raddr1 = APPLY_PREFIXING (raddr1, regs->hostregs->PX);
+        }
+#endif /*defined(_FEATURE_SIE)*/
 
 #if defined(FEATURE_EXPANDED_STORAGE)
         if(rc == 2)
         {
-	    FETCH_W(pte1,sysblk.mainstor + raddr1);
+        FETCH_W(pte1,regs->mainstor + raddr1);
             /* If page is invalid in real storage but valid in expanded
                storage then xpblk1 now contains expanded storage block# */
             if(pte1 & PAGETAB_ESVALID)
             {
-                xpblk1 = pte1 & ZPGETAB_PFRA >> 12;
+                xpblk1 = (pte1 & ZPGETAB_PFRA) >> 12;
 #if defined(_FEATURE_SIE)
                 if(regs->sie_state)
                 {
@@ -363,27 +434,23 @@ BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
                 }
 #endif /*defined(_FEATURE_SIE)*/
 
-                /* If the expanded storage block does not exist
-                   then terminate the instruction */
-                if (xpblk1 >= sysblk.xpndsize)
-                {
-                    cc = 1;
-                    goto mvpg_progck;
-                }
-
                 rc = 0;
                 xpvalid1 = 1;
-                xpkey1 = sysblk.mainstor[raddr2 +
+                xpkeya = raddr1 +
 #if defined(FEATURE_ESAME)
-                                                  2048
+                                  2048;
 #else /*!defined(FEATURE_ESAME)*/
                 /* For ESA/390 mode, the XPTE lies directly beyond 
                    the PTE, and each entry is 12 bytes long, we must
                    therefor add 1024 + 8 times the page index */
-	                             1024 + ((vaddr1 & 0x000FF000) >> 9)
+                              1024 + ((vaddr1 & 0x000FF000) >> 9);
 #endif /*!defined(FEATURE_ESAME)*/
-                                                      ];
-/*DEBUG*/ logmsg("MVPG pte1 = " F_CREG ", xkey1 = %2.2X\n",pte1,xpkey1);
+                if (xpkeya > regs->mainlim)
+                    ARCH_DEP(program_interrupt) (regs, PGM_ADDRESSING_EXCEPTION);
+                xpkey1 = regs->mainstor[xpkeya]; 
+
+/*DEBUG  logmsg("MVPG pte1 = " F_CREG ", xkey1 = %2.2X, xpblk1 = %5.5X, akey1 = %2.2X\n",
+                  pte1,xpkey1,xpblk1,akey1);  */
             }
             else
             {
@@ -396,15 +463,15 @@ BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
         /* Program check if operand not valid in main or expanded */
         if (rc)
         {
-	    cc = 1;
+        cc = 1;
             goto mvpg_progck;
-	}
+    }
 
         /* Program check if page protection or access-list controlled
            protection applies to the first operand */
-        if (prot)
+        if (prot || (xpvalid1 && (pte1 & PAGETAB_PROT)))
         {
-            regs->TEA = vaddr1 | stid | TEA_PROT_AP;
+            regs->TEA = vaddr1 | TEA_PROT_AP | stid;
             regs->excarid = (ACCESS_REGISTER_MODE(&regs->psw)) ? r1 : 0;
             ARCH_DEP(program_interrupt) (regs, PGM_PROTECTION_EXCEPTION);
         }
@@ -414,20 +481,43 @@ BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
 #if defined(FEATURE_EXPANDED_STORAGE)
     /* Program check if both operands are in expanded storage, or
        if first operand is in expanded storage and the destination
-       reference intention (register 0 bit 22) is set to one */
+       reference intention (register 0 bit 22) is set to one, or
+       if first operand is in expanded storage and pte lock bit on, or
+       if first operand is in expanded storage and frame invalid */
     if ((xpvalid1 && xpvalid2)
-        || (xpvalid1 && (regs->GR_L(0) & 0x00000200)))
+        || (xpvalid1 && (regs->GR_L(0) & 0x00000200))
+        || (xpvalid1 && (pte1 & PAGETAB_PGLOCK))
+        || (xpvalid1 && (xpblk1 >= sysblk.xpndsize)))
     {
         xcode = PGM_PAGE_TRANSLATION_EXCEPTION;
+        rc = 2;
+        cc = 1;
+        goto mvpg_progck;
+    }
+    /* More Program check checking, but at lower priority:
+       if second operand is in expanded storage and pte lock bit on, or
+       if second operand is in expanded storage and frame invalid */
+    if ((xpvalid2 && (pte2 & PAGETAB_PGLOCK))
+        || (xpvalid2 && (xpblk2 >= sysblk.xpndsize)))
+    {
+        /* re-do translation to set up TEA */
+        rc = ARCH_DEP(translate_addr) (vaddr2, r2, regs, ACCTYPE_READ, &raddr2,
+                        &xcode, &priv, &prot, &stid);
+        xcode = PGM_PAGE_TRANSLATION_EXCEPTION;
+        cc = 1;
         goto mvpg_progck;
     }
 
-    /* Perform addressing and protection checks */
+    /* Perform protection checks */
     if (xpvalid1)
     {
-        /* Perform protection check on expanded storage block */
-        if (akey1 != 0 && akey1 != (xpkey1 & STORKEY_KEY))
+        /* Key check on expanded storage block if NoKey bit off in PTE */
+        if (akey1 != 0 && akey1 != (xpkey1 & STORKEY_KEY)
+            && (pte1 & PAGETAB_ESNK) == 0
+            && !((regs->CR(0) & CR0_STORE_OVRD) && ((xpkey1 & STORKEY_KEY) == 0x90)))
+        {
             ARCH_DEP(program_interrupt) (regs, PGM_PROTECTION_EXCEPTION);
+        }
     }
     else
 #endif /*defined(FEATURE_EXPANDED_STORAGE)*/
@@ -441,10 +531,13 @@ BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
 #if defined(FEATURE_EXPANDED_STORAGE)
     if (xpvalid2)
     {
-        /* Perform protection check on expanded storage block */
+        /* Key check on expanded storage block if NoKey bit off in PTE */
         if (akey2 != 0 && (xpkey2 & STORKEY_FETCH)
-            && akey2 != (xpkey2 & STORKEY_KEY))
+            && akey2 != (xpkey2 & STORKEY_KEY)
+            && (pte2 & PAGETAB_ESNK) == 0)
+        {
             ARCH_DEP(program_interrupt) (regs, PGM_PROTECTION_EXCEPTION);
+        }
     }
     else
 #endif /*defined(FEATURE_EXPANDED_STORAGE)*/
@@ -462,40 +555,40 @@ BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
     if (xpvalid2)
     {
         /* Set the main storage reference and change bits */
-        STORAGE_KEY(aaddr1) |= (STORKEY_REF | STORKEY_CHANGE);
+        STORAGE_KEY(aaddr1, regs) |= (STORKEY_REF | STORKEY_CHANGE);
 
-	/* Set Expanded Storage reference bit in the PTE */
-        STORE_W(sysblk.mainstor + raddr2, pte2 | PAGETAB_ESREF);
+    /* Set Expanded Storage reference bit in the PTE */
+        STORE_W(regs->mainstor + raddr2, pte2 | PAGETAB_ESREF);
         
 
         /* Move 4K bytes from expanded storage to main storage */
-        memcpy (sysblk.mainstor + aaddr1,
+        memcpy (regs->mainstor + aaddr1,
                 sysblk.xpndstor + (xpblk2 << XSTORE_PAGESHIFT),
                 XSTORE_PAGESIZE);
     }
     else if (xpvalid1)
     {
         /* Set the main storage reference bit */
-        STORAGE_KEY(aaddr2) |= STORKEY_REF;
+        STORAGE_KEY(aaddr2, regs) |= STORKEY_REF;
 
-	/* Set Expanded Storage reference and change bits in the PTE */
-        STORE_W(sysblk.mainstor + raddr1, pte1 | PAGETAB_ESREF | PAGETAB_ESCHA);
+    /* Set Expanded Storage reference and change bits in the PTE */
+        STORE_W(regs->mainstor + raddr1, pte1 | PAGETAB_ESREF | PAGETAB_ESCHA);
 
         /* Move 4K bytes from main storage to expanded storage */
         memcpy (sysblk.xpndstor + (xpblk1 << XSTORE_PAGESHIFT),
-                sysblk.mainstor + aaddr2,
+                regs->mainstor + aaddr2,
                 XSTORE_PAGESIZE);
     }
     else
 #endif /*defined(FEATURE_EXPANDED_STORAGE)*/
     {
         /* Set the main storage reference and change bits */
-        STORAGE_KEY(aaddr1) |= (STORKEY_REF | STORKEY_CHANGE);
-        STORAGE_KEY(aaddr2) |= STORKEY_REF;
+        STORAGE_KEY(aaddr1, regs) |= (STORKEY_REF | STORKEY_CHANGE);
+        STORAGE_KEY(aaddr2, regs) |= STORKEY_REF;
 
         /* Move 4K bytes from main storage to main storage */
-        memcpy (sysblk.mainstor + aaddr1,
-                sysblk.mainstor + aaddr2,
+        memcpy (regs->mainstor + aaddr1,
+                regs->mainstor + aaddr2,
                 XSTORE_PAGESIZE);
     }
 
@@ -504,6 +597,7 @@ BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
     return;
 
 mvpg_progck:
+
     /* If page translation exception (PTE invalid) and condition code
         option in register 0 bit 23 is set, return condition code */
     if ((regs->GR_L(0) & 0x00000100)
@@ -529,11 +623,15 @@ mvpg_progck:
 
 #if !defined(_GEN_ARCH)
 
-#define  _GEN_ARCH 390
-#include "xstore.c"
+#if defined(_ARCHMODE2)
+ #define  _GEN_ARCH _ARCHMODE2
+ #include "xstore.c"
+#endif
 
-#undef   _GEN_ARCH
-#define  _GEN_ARCH 370
-#include "xstore.c"
+#if defined(_ARCHMODE3)
+ #undef   _GEN_ARCH
+ #define  _GEN_ARCH _ARCHMODE3
+ #include "xstore.c"
+#endif
 
 #endif /*!defined(_GEN_ARCH)*/
